<?php

/**
 * @file Contains Drupal\force_password_change\Service\ForcePasswordChangeService
 *
 * Maps database requests for data regarding friendships
 */

namespace Drupal\force_password_change\Service;

use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Query\Condition;
use Drupal\Core\Session\AccountProxy;
use Drupal\user\UserData;

class ForcePasswordChangeService implements ForcePasswordChangeServiceInterface
{
	/**
	 * The database connection
	 *
	 * @var \Drupal\Core\Database\Connection
	 */
	protected $connection;

	/**
	 * The current user
	 *
	 * @var \Drupal\Core\Session\AccountProxy
	 */
	protected $currentUser;

	/**
	 * The config factory object
	 *
	 * @var \Drupal\Core\Config\ConfigFactory
	 */
	protected $configFactory;

	/**
	 * The config factory object
	 *
	 * @var \Drupal\user\UserData
	 */
	protected $userData;

	/**
	 * Constructs a ForcePasswordChangeService object.
	 *
	 * @param \Drupal\Core\Database\Connection $connection
	 *   The database connection
	 * @param \Drupal\Core\Session\AccountProxy $currentUser
	 *   The current user
	 * @param \Drupal\Core\Config\ConfigFactory $configFactory
	 *   The config factory
	 * @param \Drupal\user\UserData $userData
	 *   The user data service
	 */
	public function __construct(Connection $connection, AccountProxy $currentUser, ConfigFactory $configFactory, UserData $userData)
	{
	  $this->connection = $connection;
	  $this->currentUser = $currentUser;
	  $this->configFactory = $configFactory;
	  $this->userData = $userData;
	}

	public function forceUsersPasswordChange($uids = array())
	{
		if(!count($uids))
		{
			$uids = db_query('SELECT uid FROM {users_field_data} WHERE status = :one', [':one' => 1])->fetchCol();
		}

		foreach($uids as $uid)
		{
			$this->userData->set('force_password_change', $uid, 'pending_force', 1);
			$this->userData->set('force_password_change', $uid, 'last_force', REQUEST_TIME);
		}
	}

	public function checkForForce()
	{
		// Default is to not redirect
		$redirect = FALSE;

		// If the user's account has been flagged for expiry, a redirect is required
		if($this->userData->get('force_password_change', $this->currentUser->id(), 'pending_force'))
		{
			$redirect = 'admin_forced';
		}
		// Only check to see if their password has expired if password expiry
		// is turned on in the module settings page.
		elseif($this->configFactory->get('force_password_change.expire_password')->get('expire'))
		{
			// The user's account has not been flagged for password expiry. Check to see
			// if their password has expired according to the rules of the module.

			// First thing is to check the time of their last password change,
			// and the time of their account creation
			$last_change = $this->userData->get('force_password_change', $this->currentUser->id(), 'last_change');
			$created = $this->connection->select('users_field_data', 'ufd')
				->fields('ufd', array('created'))
				->condition('ufd.uid', $this->currentUser->id())
				->execute()
				->fetchField();

			// Get the time period after which their password should expire
			// according to the rules laid out in the module settings page. Only the
			// role with the highest priority is retrieved
			$query = $this->connection->select('force_password_change_expiry', 'fpce');
			$expiry = $query
				->fields('fpce', array('expiry'))
				->condition('fpce.rid', array_keys($this->currentUser->getRoles()), 'IN')
				->orderBy('fpce.weight')
				->range(0, 1)
				->addTag('force_password_change_expiry_check')
				->execute()
				->fetchField();

			// Test to see if their password has expired
			if($expiry && ($last_change != '' && (REQUEST_TIME - $expiry) > $last_change) || ($last_change == '' && (REQUEST_TIME - $expiry) > $created))
			{
				// Their password has expired, so their user account is flagged
				// and the expiration time period is returned, which will trigger the redirect
				// and be used to generate the message shown to the user
				$this->userData('force_password_change', $this->currentUser->id(), 'pending_force', 1);

				$redirect = 'expired';
			}
		}

		return $redirect;
	}

	public function getLastChangeForRole($rid)
	{
		return $this->connection->query('SELECT last_force FROM {force_password_change_roles} WHERE rid = :rid', array(':rid' => $rid))->fetchField();
	}

	public function updateLastChangeForRoles($rids)
	{
		$query = $this->connection->update('force_password_change_roles')
			->fields(['last_force' => REQUEST_TIME])
			->condition('rid', $rids, 'IN')
			->execute();
	}

	public function insertExpiryForRoles($values)
	{
		if(count($values))
		{
			// Prepare the insert query for new roles that have not had their password expiry set
			$query = $this->connection->insert('force_password_change_expiry')
				->fields(['rid', 'expiry', 'weight']);
			foreach($values as $role_values)
			{
				$query->values($role_values);
			}

			$query->execute();
		}
	}

	public function updateExpiryForRole($rid, $time_period, $weight)
	{
		$update_query = $this->connection->update('force_password_change_expiry')
			->fields(['expiry' => $time_period, 'weight' => $weight])
			->condition('rid', $rid)
			->execute();
	}

	public function getUserCountForRole($rid)
	{
		if($rid == 'authenticated')
		{
			$count = $this->connection->query('SELECT COUNT(uid) FROM {users_field_data} WHERE status = 1')->fetchField();
		}
		else
		{
			$count = $this->connection->query
			(
				'SELECT COUNT(data.uid) ' .
				'FROM {users_field_data} AS data ' .
				'LEFT JOIN {user__roles} AS roles ' .
				'ON roles.entity_id = data.uid AND roles.roles_target_id = :rid ' .
				'WHERE data.status = 1',
				[
					':rid' => $rid,
				]
			)->fetchField();
		}

		return $count;
	}

	public function getPendingUsersForRole($rid, $countQuery = FALSE)
	{
		$pending_query = $this->connection->select('users_data', 'ud')->extend('Drupal\Core\Database\Query\PagerSelectExtender');
		$pending_query->addField('ud', 'uid');
		$pending_query->addTag('force_password_change_pending_users')
			->limit(20)
			->condition('ud.module', 'force_password_change')
			->condition('ud.name', 'pending_force')
			->condition('ud.value', 1);

		// If the role is anything other than the authenticated users role, we need to
		// limit the users to the members of that role
		if($rid != 'authenticated')
		{
			$alias2 = $pending_query->join('user__roles', 'ur', 'ur.entity_id = ud.uid');
			$pending_query->condition($alias2 . '.roles_target_id', $rid);
		}
		$pending_user_ids = $pending_query->execute()->fetchCol();

		if($countQuery)
		{
			return count($pending_user_ids);
		}

		return user_load_multiple($pending_user_ids);
	}

	public function getNonPendingUsersForRole($rid)
	{
		$nonpending_query = $this->connection->select('users', 'u')->extend('Drupal\Core\Database\Query\PagerSelectExtender');
		$alias = $nonpending_query->leftJoin('users_data', 'ud', 'ud.uid = u.uid AND ud.module = :force_password_change AND ud.name = :pending_force AND ud.value = :one', [':force_password_change' => 'force_password_change', ':pending_force' => 'pending_force', ':one' => 1]);
		$nonpending_query->addField('u', 'uid');
		$nonpending_query
			->addTag('force_password_change_pending_users')
			->limit(100)
			->condition('u.uid', 0, '!=')
			->isNull('ud.uid');

		if($rid != 'authenticated')
		{
			$alias2 = $nonpending_query->join('user__roles', 'ur', 'ur.entity_id = ud.uid');
			$nonpending_query->condition($alias2 . '.roles_target_id', $rid);
		}

		$nonpending_user_ids = $nonpending_query->execute()->fetchCol();

		return user_load_multiple($nonpending_user_ids);
	}

	public function getRoleExpiryDates()
	{
		return $this->connection->query('SELECT rid, expiry, weight from {force_password_change_expiry} ORDER BY weight, rid');
	}

	public function getUsersForRole($rid, $uidOnly = TRUE)
	{
		$uids = db_query
		(
			'SELECT entity_id ' .
			'FROM {user__roles} ' .
			'WHERE roles_target_id = :rid',
			[':rid' => $rid]
		)->fetchCol();

		if($uidOnly)
		{
			return $uids;
		}

		return user_load_multiple($uids);
	}

	public function setChangedTime($uid)
	{
		$this->userData->set('force_password_change', $uid, 'last_change', REQUEST_TIME);
	}

	public function removePendingForce($uid)
	{
		$this->userData->set('force_password_change', $uid, 'pending_force', 0);
	}

	/**
	 * Converts a time period of password expiry dates to the textual representation
	 * of the number of years, months, days or hours that it represents
	 */
	public function getTextDate($timestamp)
	{
		$year = 60 * 60 * 24 * 365;
		if($timestamp % $year === 0)
		{
			$time_period = $timestamp / $year;
			$time_period = ($time_period > 1) ? $time_period . ' ' . t('years') : t('year');
		}
		else
		{
			$week = 60 * 60 * 24 * 7;
			if($timestamp % $week === 0)
			{
				$time_period = $timestamp / $week;
				$time_period = ($time_period > 1) ? $time_period . ' ' . t('weeks') : t('week');
			}
			else
			{
				$day = 60 * 60 * 24;
				if($timestamp % $day === 0)
				{
					$time_period = $timestamp / $day;
					$time_period = ($time_period > 1) ? $time_period . ' ' . t('days') : t('day');
				}
				else
				{
					$hour = 60 * 60;
					if($timestamp % $hour === 0)
					{
						$time_period = $timestamp / $hour;
						$time_period = ($time_period > 1) ? $time_period . ' ' . t('hours') : t('hour');
					}
				}
			}
		}

		return $time_period;
	}
}
